""" ccaaws cloudformation class
"""
import sys
import time

from botocore.exceptions import ClientError
import ccalogging
from ccautils.errors import errorRaise
import ccautils.fileutils as FT
import ccautils.utils as UT

from ccaaws import __version__
from ccaaws.botosession import BotoSession

log = ccalogging.log


class CFNClient(BotoSession):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.newClient("cloudformation")

    def stackDetails(self, stackname):
        try:
            resp = self.client.describe_stacks(StackName=stackname)
            if "Stacks" in resp:
                stack = resp["Stacks"][0]
                return stack
        except ClientError:
            log.debug(f"stack: {stackname} does not exist")
            return None
        except Exception as e:
            fname = sys._getframe().f_code.co_name
            errorRaise(fname, e)

    def stackStatus(self, stackname):
        try:
            stack = self.stackDetails(stackname)
            return stack["StackStatus"] if stack is not None else None
        except Exception as e:
            fname = sys._getframe().f_code.co_name
            errorRaise(fname, e)

    def createStack(self, **kwargs):
        try:
            resp = self.client.create_stack(**kwargs)
            if "StackId" in resp:
                return resp["StackId"]
            return None
        except Exception as e:
            fname = sys._getframe().f_code.co_name
            errorRaise(fname, e)

    def updateStack(self, **kwargs):
        try:
            resp = self.client.update_stack(**kwargs)
            if "StackId" in resp:
                return resp["StackId"]
            return None
        except Exception as e:
            fname = sys._getframe().f_code.co_name
            errorRaise(fname, e)

    def deleteStack(self, stackname):
        try:
            self.client.delete_stack(StackName=stackname)
        except Exception as e:
            fname = sys._getframe().f_code.co_name
            errorRaise(fname, e)

    def waitForStack(self, stackn, timeout=10, sleeptime=30):
        """ wait for a stack to become "..._COMPLETE"

        waits for timeout * sleeptime seconds

        returns the stack status
        """
        try:
            status = None
            cn = 0
            sleeptime = sleeptime
            while True:
                cn += 1
                if cn > timeout:
                    log.error(
                        f"Timeout expired waiting for stack {stackn} to become ready"
                    )
                    break
                status = self.stackStatus(stackn)
                if status is not None and "COMPLETE" in status:
                    log.info(f"Stack {stackn} is {status}")
                    break
                elif status is None:
                    log.warning(f"stack {stackn} does not exist (anymore)")
                    break
                time.sleep(sleeptime)
            return status
        except Exception as e:
            fname = sys._getframe().f_code.co_name
            errorRaise(fname, e)

    def checkStack(self, stack):
        ok = False
        log.debug(f"checking stack: {stack}")
        if "StackStatus" in stack:
            log.debug("status is in stack")
            if stack["StackStatus"] in ["UPDATE_COMPLETE", "CREATE_COMPLETE"]:
                log.debug("is one of the updates")
                if "Tags" in stack:
                    log.debug("stack has tags")
                    for tag in stack["Tags"]:
                        if tag["Key"] == "version":
                            log.debug(
                                f"""version tag: {tag["Value"]}, version: {__version__}"""
                            )
                            if tag["Value"] == __version__:
                                log.debug("version tag is correct")
                                ok = True
        return ok

    def expandDictToList(self, xdict, keyname="Key", valuename="Value"):
        """ expands a dictionary into a list of dictionaries, one per key

            {"somekey": "someval", "otherkey": "otherval"}

        returns a list of dictionaries

            [
                {"Key": "somekey", "Value": "someval"},
                {"Key": "otherkey", "Value": "otherval"},
            ]
        """
        try:
            op = []
            for key in xdict:
                tmp = {keyname: key, valuename: xdict[key]}
                op.append(tmp)
            return op
        except Exception as e:
            fname = sys._getframe().f_code.co_name
            errorRaise(fname, e)

    def buildStackTags(self, tagdict):
        """ converts a set of tags into the form used by CFN Stacks

                tagdict = {
                    "owner": "sre",
                    "environment": "prod",
                    "product": "ccaaws",
                }

        returns a list of dictionaries
            [
                {"Key": "owner", "Value": "sre"},
                {"Key": "environment", "Value": "prod"},
                {"Key": "product", "Value": "ccaaws"},
            ]
        """
        try:
            tagdict["version"] = __version__
            tagdict["builder"] = "ccaaws"
            return self.expandDictToList(tagdict)
        except Exception as e:
            fname = sys._getframe().f_code.co_name
            errorRaise(fname, e)

    def buildStackParams(self, paramdict):
        """ builds a set of parameters from a paramstring
        """
        try:
            lpd = self.expandDictToList(
                paramdict, keyname="ParameterKey", valuename="ParameterValue"
            )
            return lpd
        except Exception as e:
            fname = sys._getframe().f_code.co_name
            errorRaise(fname, e)

    def testParam(
        self,
        param,
        options,
        testlist=False,
        teststr=False,
        testint=False,
        testdict=False,
    ):
        if param in options and options[param] is not None:
            if testlist:
                return True if type(options[param]) is list else False
            elif teststr:
                return True if type(options[param]) is str else False
            elif testint:
                return True if type(options[param]) is int else False
            elif testdict:
                return True if type(options[param]) is dict else False
            else:
                return True
        return False

    def buildStackDict(self, options):
        """ sets up the stack 'dictionary' for the create/update stack functions

        param: options
               a dictionary of required options for the stack
                "templatefn": fqfn of the yaml or json template for the stack
                "stackname": the name of the stack
                "params"(optional): the parameters of the stack in string form
                          'param1=val1,param2=val2'
                    or dict form:
                          {"param1": "val1", "param2": "val2"}
                "tags"(optional): the stack tags in dict form
                "capabilities"(optional): ["CAPABILITY_IAM"] or ["CAPABILITY_NAMED_IAM"]

        returns a dictionary ready for the stack create/update functions
        """
        try:
            template = FT.readFile(options["templatefn"])
            if template is None:
                raise Exception(f"""Failed to read {options["templatefn"]}""")
            xd = {"StackName": options["stackname"], "TemplateBody": template}
            if self.testParam("params", options, teststr=True):
                pd = UT.makeDictFromString(options["params"])
                xd["Parameters"] = self.buildStackParams(pd)
            elif self.testParam("params", options, testdict=True):
                xd["Parameters"] = self.buildStackParams(options["params"])
            if self.testParam("tags", options, testdict=True):
                xd["Tags"] = self.buildStackTags(options["tags"])
            else:
                xd["Tags"] = self.buildStackTags({})
            if self.testParam("capabilities", options, testlist=True):
                if len(options["capabilities"]) > 0:
                    xd["Capabilities"] = options["capabilities"]
            return xd
        except Exception as e:
            fname = sys._getframe().f_code.co_name
            errorRaise(fname, e)

    def installOrUpdateStack(self, options):
        """ installs a stack if it doesn't exist or updates it

        param options: as for the buildStackDict function above

        returns stack status or None
        """
        try:
            status = self.waitForStack(options["stackname"])
            xd = self.buildStackDict(options)
            if status is not None and "COMPLETE" in status:
                if not self.checkStack(self.stackDetails(options["stackname"])):
                    # stack exists, so update it
                    log.warning(f"""updating stack {options["stackname"]}""")
                    self.updateStack(**xd)
                    time.sleep(10)
                    status = self.waitForStack(options["stackname"])
                else:
                    log.info(f"""Stack {options["stackname"]} is up to date""")
            elif status is None:
                log.info(f"""creating stack {options["stackname"]}""")
                self.createStack(**xd)
                time.sleep(10)
                status = self.waitForStack(options["stackname"])
            else:
                msg = f"""stack {options["stackname"]} is status: {status}"""
                log.warning(msg)
                raise Exception(msg)
        except ClientError as ce:
            log.warning("Client Error: stack probably already up to date")
            log.warning(f"{ce}")
        except Exception as e:
            fname = sys._getframe().f_code.co_name
            errorRaise(fname, e)



import sys

from ccautils.errors import errorRaise
import ccautils.utils as UT

from ccaaws.botosession import BotoSession


class ACM(BotoSession):
    def __init__(self, **kwargs):
        try:
            if kwargs is None:
                kwargs = {"noresource": True}
            else:
                kwargs["noresource"] = True
            super().__init__(**kwargs)
            self.newClient("acm")
        except Exception as e:
            fname = sys._getframe().f_code.co_name
            errorRaise(fname, e)

    def getCertList(self):
        try:
            pag = self.client.get_paginator("list_certificates")
            pageit = pag.paginate()
            certs = []
            for page in pageit:
                if "CertificateSummaryList" in page:
                    for cert in page["CertificateSummaryList"]:
                        certs.append(cert["CertificateArn"])
            return certs
        except Exception as e:
            fname = sys._getframe().f_code.co_name
            errorRaise(fname, e)

    def getCertDetails(self, certarn):
        try:
            cert = self.client.describe_certificate(CertificateArn=certarn)
            return cert["Certificate"]
        except Exception as e:
            fname = sys._getframe().f_code.co_name
            errorRaise(fname, e)

    def _setupTCert(self, arn, cert):
        tcert = None
        isemail = False
        if len(cert) > 0:
            # print("{}".format(cert))
            tcert = {"arn": arn}
            tcert["inuse"] = []
            if "InUseBy" in cert:
                tcert["inuse"] = cert["InUseBy"]
            if "NotAfter" in cert:
                tcert["expirests"], tcert["expires"] = UT.fuzzyExpires(cert["NotAfter"])
            else:
                tcert["expirests"] = 0
                tcert["expires"] = ""
            tcert["domain"] = cert["DomainName"]
            tcert["altdomains"] = cert["SubjectAlternativeNames"]
            vopt = []
            tcert["status"] = "unknown"
            for opt in cert["DomainValidationOptions"]:
                tcert["validation"] = None
                try:
                    if "ValidationMethod" in opt:
                        tcert["validation"] = opt["ValidationMethod"]
                    if tcert["validation"] == "EMAIL":
                        isemail = True
                    if "ValidationStatus" in opt:
                        tcert["status"] = opt["ValidationStatus"]
                    xval = {
                        "domain": opt["DomainName"],
                        "validationdomain": opt["ValidationDomain"],
                    }
                    vopt.append(xval)
                except KeyError:
                    continue
            tcert["validations"] = vopt
        return (isemail, tcert)

    def getCerts(self, emailvalidationonly=True):
        try:
            carns = self.getCertList()
            # print(carns)
            certs = []
            ecn = 0
            dcn = 0
            for arn in carns:
                cert = self.getCertDetails(arn)
                isemail, tcert = self._setupTCert(arn, cert)
                if tcert is not None:
                    if isemail:
                        ecn += 1
                    else:
                        dcn += 1
                    if emailvalidationonly:
                        if tcert["validation"] == "EMAIL":
                            certs.append(tcert)
                    else:
                        certs.append(tcert)
            return (ecn, dcn, certs)
        except Exception as e:
            fname = sys._getframe().f_code.co_name
            errorRaise(fname, e)



"""AWS Dynamo DB client."""

import ccalogging

from ccaaws.botosession import BotoSession

log = ccalogging.log


class DynamoDB(BotoSession):

    LCACHE = []

    def __init__(self, **kwargs):
        self.dolcache = False
        if kwargs is not None and "LambdaCache" in kwargs:
            self.dolcache = kwargs["LambdaCache"]
            del kwargs["LambdaCache"]
        super().__init__(**kwargs)
        self.newClient("dynamodb")

    def getAll(self, table):
        if self.dolcache and len(self.LCACHE) > 0:
            return self.LCACHE
        resp = self.client.scan(TableName=table)
        data = resp["Items"]
        while resp.get("LastEvaluatedKey"):
            resp = self.client.scan(
                TableName=table, ExclusiveStartKey=resp["LastEvaluatedKey"]
            )
            data.extend(resp["Items"])
        if self.dolcache:
            self.LCACHE = data
        return data

    def getItem(self, item):
        """Item is a dictionary in dynamodb stylee.

        see:
        https://boto3.readthedocs.io/en/latest/
        reference/services/dynamodb.html#DynamoDB.Client.get_item

        i.e.
        item={
          "TableName"='tablename',
          "Key"={
            'accountname':{
              'S':'connectedhomes-dev',
            },
          }
        }

        Args:
            item: item dictionary
        """
        resp = None
        try:
            resp = self.client.get_item(TableName=item["TableName"], Key=item["Key"])
        except Exception as e:
            log.warning(f"Failed to retrieve item {item}, Exception was {e}")
        return resp

    def putItem(self, item):
        """
        item is a dict in dynamodb style, see link above and check put_item
        """
        resp = None
        try:
            resp = self.client.put_item(TableName=item["TableName"], Item=item["Item"])
        except Exception as e:
            log.warning(f"Failed to put item {item}, Exception was {e}")
        return resp

    def deleteItem(self, item):
        try:
            self.client.delete_item(TableName=item["TableName"], Key=item["Key"])
            log.info(f"deleted: {item}")
        except Exception:
            log.warning(f"Failed to delete {item}")



"""S3 filesystem object."""

import os

import ccalogging

from ccaaws.botosession import BotoSession

log = ccalogging.log


class S3FileSystem(BotoSession):
    """Class to mimic filesystem tools with S3."""

    def __init__(self, **kwargs):
        """Connect to AWS S3.

        Optional:
            region: region name
            profile: profile name
        """
        super().__init__(**kwargs)
        self.newClient("s3")
        self.bucket = None

    def getBuckets(self):
        try:
            resp = self.client.list_buckets()
            buckets = resp["Buckets"]
            return buckets
        except KeyError as e:
            log.error(f"Failed to retrieve the bucket list: {e}")

    def xls(self, path, page_size=None):
        if path == "":
            return [[], []]
        paginator = self.client.get_paginator("list_objects")
        paging_args = {
            "Bucket": self.bucket,
            "Prefix": path,
            "Delimiter": "/",
            "PaginationConfig": {"PageSize": page_size},
        }
        iterator = paginator.paginate(**paging_args)
        objs = []
        paths = []
        for resp in iterator:
            # print(resp)
            xpaths = resp.get("CommonPrefixes", [])
            xobjs = resp.get("Contents", [])
            for xobj in xobjs:
                objs.append(xobj)
            for xpath in xpaths:
                paths.append(xpath)
        return [objs, paths]

    def xcp(self, src, dest):
        """Copy to/from S3.

        Args:
            src: source path
            dest: destination path
        """
        if src.startswith("s3://"):
            ret = self._cpFromS3(src, dest)
        else:
            ret = self._cpToS3(src, dest)
        return ret

    def _cpFromS3(self, src, dest):
        """Copy from S3.

        Args:
            src: source path
            dest: destination path
        """
        ret = False
        scheme, bucket, path = self.parseS3Uri(src)
        self.bucket = bucket
        objs, paths = self.xls(path)
        sz = len(objs)
        if sz == 1:
            try:
                # path="/{}".format(path)
                log.debug(f"copying {bucket} {path} to {dest}")
                self.resource.meta.client.download_file(bucket, path, dest)
                ret = True
            except Exception as E:
                log.warning(f"failed to copy file {src} to {dest}, exception was: {E}")
        elif sz == 0:
            log.warning(f"source file not found in s3 {src}")
        else:
            log.warning(f"more than one object returned for {src}. Objects: {objs}")
        return ret

    def _cpToS3(self, src, dest):
        """Copies src to s3://dest.

        Args:
            src: the source path
            dest: the destination path
        """
        ret = False
        if os.path.isfile(src):
            scheme, bucket, path = self.parseS3Uri(dest)
            try:
                self.resource.meta.client.upload_file(src, bucket, path)
                ret = True
            except Exception as E:
                log.warning(f"failed to copy file {src} to {dest}, exception was: {E}")
        else:
            log.warning(f"source file not found {src}")
        return ret

    def parseS3Uri(self, uri):
        """Parses S3 uris.

        s3://bucketname/some/path/some.file
        or
        http(s)://bucketname.s3.amazonaws.com/some/path/some.file

        into a tuple of scheme, bucket and path

        Args:
            uri: the uri to parse
        """
        path = ""
        scheme, bucket = self.parseBucket(uri)
        lscheme = len(scheme)
        if lscheme > 0:
            slashpos = uri.find("/", lscheme + 3)
            if slashpos > -1:
                path = uri[slashpos + 1 :]
        return [scheme, bucket, path]

    def parseScheme(self, uri):
        """Returns the scheme from the supplied uri.

        Args:
            uri: the fully qualified path
                 s3://bucket/file/name
                 https://s3.amazonaws.com/bucket/file/name
        """
        scheme = ""
        schemepos = uri.find("://")
        if schemepos > -1:
            scheme = uri[:schemepos]
        return scheme

    def parseBucket(self, uri):
        """Returns the bucket part of the supplied uri.

        Args:
            uri: the fully qualified path
                 s3://bucket/file/name
                 https://s3.amazonaws.com/bucket/file/name
        """
        bucket = ""
        scheme = self.parseScheme(uri)
        lscheme = len(scheme)
        if lscheme > 0:
            bucketpos = lscheme + 3
            if scheme.startswith("http"):
                dotpos = uri.find(".")
                if dotpos > -1:
                    bucket = uri[bucketpos:dotpos]
            else:
                # s3:// uri
                slashpos = uri.find("/", bucketpos)
                if slashpos > -1:
                    bucket = uri[bucketpos:slashpos]
                else:
                    # just a bucket name, then, with no trailing path
                    bucket = uri[bucketpos:]
        return [scheme, bucket]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ccaaws']

package_data = \
{'': ['*']}

install_requires = \
['boto3>=1.11.5,<2.0.0', 'ccalogging>=0.3.3,<0.4.0', 'ccautils>=0.4.0,<0.5.0']

setup_kwargs = {
    'name': 'ccaaws',
    'version': '0.4.8',
    'description': 'AWS boto3 objects for python',
    'long_description': '# ccaaws\n\nPython objects wrapped around [boto3](https://github.com/boto/boto3)\n\n## botosession.py\n\nBase class inherited by all the others.  Manages connections to AWS via\naccess keys or profiles.\n\n## cfnclient.py\n\nCloudFormation client\n\n## cwclient.py\n\nCloudWatch client\n\n## iamclient.py\n\nAccess to IAM functions - mainly for user access key management.\n\nCan raise an `AccessKeyError` exception on failure.\n\n```\nfrom ccaaws.iamclient import IamClient\n\niam = IamClient("iam.user")\nnewkey = iam.rotateKeys()\n```\n\n## paramstore.py\n\nSSM ParameterStore client\n\n[modeline]: # ( vim: set ft=markdown tw=74 fenc=utf-8 spell spl=en_gb mousemodel=popup: )\n',
    'author': 'ccdale',
    'author_email': 'chris.charles.allison+ccaaws@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



""" ccaaws ec2 class
"""
import sys

from botocore.exceptions import ClientError
import ccalogging
from ccautils.errors import errorRaise

from ccaaws.botosession import BotoSession

log = ccalogging.log


class EC2(BotoSession):
    def __init__(self, **kwargs):
        """ connect to aws

        to switch region supply the `region` keyword
        """
        super().__init__(**kwargs)
        self.newClient("ec2")

    def findInstances(self, instlist):
        """ find instances in instlist

        instlist is a list of instance-id strings
        """
        try:
            instances = []
            kwargs = {"InstanceIds": instlist} if type(instlist) is list else {}
            while True:
                try:
                    # will raise client error if instances don't exist
                    resp = self.client.describe_instances(**kwargs)
                    try:
                        rinsts = [
                            i["Instances"]
                            for i in [
                                r for r in resp["Reservations"] if "Instances" in r
                            ]
                        ]
                        instances += [i for subi in rinsts for i in subi]
                        kwargs["NextToken"] = resp["NextToken"]
                    except KeyError:
                        break
                except ClientError as ce:
                    log.debug(f"ClientError: Instances probably don't exist: {ce}")
                    break
            return instances
        except Exception as e:
            fname = sys._getframe().f_code.co_name
            errorRaise(fname, e)

    def getMatchingInstances(self, instlst=None):
        """ find the instances named in the `instlst` list
        or all instances if `instlst` is None

        NOTE: this does not appear to work with filters, use
        the 'findInstances' function above
        """
        try:
            instances = []
            kwargs = {}
            if type(instlst) is list:
                kwargs["Filters"] = [{"Name": "instance-id", "Values": instlst}]
            while True:
                resp = self.client.describe_instances(**kwargs)
                try:
                    rinsts = [
                        i["Instances"]
                        for i in [r for r in resp["Reservations"] if "Instances" in r]
                    ]
                    instances += [i for subi in rinsts for i in subi]
                    kwargs["NextToken"] = resp["NextToken"]
                except KeyError:
                    break
            return instances
        except Exception as e:
            fname = sys._getframe().f_code.co_name
            errorRaise(fname, e)

    def getRegions(self):
        """
        returns a list of all available regions
        """
        try:
            resp = self.client.describe_regions()
            regions = [
                region["RegionName"] for region in resp["Regions"] if "Regions" in resp
            ]
            log.debug("Regions: {}".format(regions))
            return regions
        except Exception as e:
            fname = sys._getframe().f_code.co_name
            errorRaise(fname, e)
